import {camelCase as camelCaseUntyped} from 'camel-case'
import {snakeCase as snakeCaseUntyped} from 'snake-case'
import type {
	camelCase as camelCaseTyped,
	snakeCase as snakeCaseTyped,
	CamelCase,
	SnakeCase,
} from 'literal-case'

const camelCase = camelCaseUntyped as typeof camelCaseTyped
const snakeCase = snakeCaseUntyped as typeof snakeCaseTyped

// type ObjectToCamelCase<T> = {
// 	[P in keyof T as (P extends string
// 		? CamelCase<P>
// 		: P
// 	)]: T[P] extends (Record<string, unknown> | unknown[])
// 		? ObjectToCamelCase<T[P]>
// 		: T[P]
// }
type NestedObjectsToCamelCase<
	// T extends (Record<string, unknown> | unknown[])
	T
> = T extends unknown[]
	? {
		// Change array elements while keeping prototype properties the same.
		[P in keyof T]: P extends number
			? NestedObjectsToCamelCase<T[P]>
			: T[P]
	}
	// TODO what about class instances here? Methods would also get transformed. Is it right?
	// : T extends Record<string, any>
	: T extends Record<string, unknown>
	? {
		[P in keyof T as (P extends string
			? CamelCase<P>
			: P
		)]: NestedObjectsToCamelCase<T[P]>
	}
	: T

/**
 * Transforms field names to camel case.
 */
export function objectToCamelCase<T extends Record<string, any>>(
	object: T,
): NestedObjectsToCamelCase<T> {

	// When calling recursively, this can be called without being and object or array in which case we just return the value
	if (typeof object !== 'object') {
		return object
	}

	const parsedObject: Record<string, any> = {}
	for (const m in object) {
		parsedObject[camelCase(m)] = object[m]

		// Recursive processing
		// Prevent processing for some cases
		if (object[m] === null) {
			continue
		}

		// Call it again for arrays
		if (Array.isArray(object[m])) {
			parsedObject[camelCase(m)] = object[m].map((o: Record<string, any>) => objectToCamelCase(o))
			// Because typeof [] === 'object' is true for arrays, we leave the loop here to prevent converting arrays to objects.
			continue
		}

		// Call it again for nested objects
		if (typeof object[m] === 'object') {
			parsedObject[camelCase(m)] = objectToCamelCase(object[m])
		}
	}
	return parsedObject
}

/**
 * Transforms field names to snake case - used before making an api request.
 */
export function objectToSnakeCase(object: Record<string, any>) {

	// When calling recursively, this can be called without being and object or array in which case we just return the value
	if (typeof object !== 'object') {
		return object
	}

	const parsedObject: Record<string, any> = {}
	for (const m in object) {
		parsedObject[snakeCase(m)] = object[m]

		// Recursive processing
		// Prevent processing for some cases
		if (
			object[m] === null ||
			(object[m] instanceof Date)
		) {
			continue
		}

		// Call it again for arrays
		if (Array.isArray(object[m])) {
			parsedObject[snakeCase(m)] = object[m].map((o: Record<string, any>) => objectToSnakeCase(o))
			// Because typeof [] === 'object' is true for arrays, we leave the loop here to prevent converting arrays to objects.
			continue
		}

		// Call it again for nested objects
		if (typeof object[m] === 'object') {
			parsedObject[snakeCase(m)] = objectToSnakeCase(object[m])
		}
	}

	return parsedObject
}
